@make(ais)
@aiInfo(topic=<Boolean operators producing out of range results>,
number=[AI-00535/03],
classcode=[BI],
statuscode=[WA],
mainref=[04.05.06@\(03)],
date=[88-11-04])
@aiHeading()
@begin(history)
@begin(toplines)
@mainref()
@standard[04.05.01@\(03)]
@standard[11.07@\(10)]
!class binding interpretation 87-04-13
@cbon()!status approved by WG9 88-10-28@cboff()
@end(toplines)
!status ARG-approved 88-05-10 (reviewed)
!status ARG-approved (16-0-0) 87-09-15 (pending editorial review)
!status panel-approved (6-0-0) 87-09-15
!status work-item 87-04-15
!status received 87-04-13
@references[!references @CommentNum[83-00915]83-00915]
@topic{!topic Boolean operators producing out of range results}
@end(history)

@subheading[!summary 87-04-13]

When the predefined @b[not] operator or a predefined logical operator delivers a
result having an array type, CONSTRAINT_@|ERROR is raised if any component of
the result does not belong to the component subtype.  The check corresponding
to this exception situation is included in the set of checks defined for
RANGE_@|CHECK.

@subheading[!question 87-04-15]

Consider the following example:
@begin(example)
@key[procedure] BOOLSUB @key[is]
    @key[subtype] T_@|BOOL @key[is] BOOLEAN @key[range] TRUE..TRUE;
    @key[type] ARR @key[is] @key[array] (1..10) @key[of] T_@|BOOL;
    A, B : ARR := (1..10 => TRUE); -- the only allowed value
@key[begin]
    A := @key[not] A;
-- Here A(1)=A(2)=...=A(10)= FALSE, but no exception is raised
    A := B @key[xor] B;
-- Similarly, all components of A are FALSE
@key[end] BOOLSUB;
@end(example)
The program is legal, and there is no rule in the Standard that requires an
exception to be raised either by the @b[not] operator, the @b[xor] operator, or the
assignment operation.  Was it the intent that an exception be raised under
these circumstances?  (Note: this peculiar situation can arise only for the
@b[not] and @b[xor] operators applied to arrays.)

@subheading[!recommendation 87-04-15]

When the predefined @b[not] operator or a predefined logical operator delivers a
result having an array type, CONSTRAINT_@|ERROR is raised if any component of
the result does not belong to the component subtype.  The check corresponding
to this exception situation is included in the set of checks defined for
RANGE_@|CHECK.

@subheading[!discussion 88-03-30]

A principal goal of Ada's design was to ensure that no initialized object ever
holds a value that violates its constraint.  This goal is clearly not achieved
for the @b[not] and @b[xor] operators when they are applied to arrays such as
those given in the example.  To ensure that the design intent is realized, the
result delivered by @b[not] and @b[xor] must be checked to ensure each
component of the result satisfies the component constraint of the array type.
Note that no such checks are needed for the @b[and] and @b[or] operators,
although it doesn't hurt to formulate a general rule that includes these
operators.  Also, the checks cannot be performed as part of the assignment
operation because this operation assumes that component values satisfy their
constraints, so no check is needed.

The only predefined operators that produce array results are the logical
operators, @b[not], and the catenation operator.  Since catenation does not
produce new component values, the problem illustrated by the example can only
arise for @b[not] and the logical operators.

Imposing the recommended rule will not affect code efficiency when an array
component has an unconstrained boolean subtype since the check can never fail
in this case.  It is only when the component subtype is constrained to a
single value that special code is needed.  If it is known at compile-time
that the component subtype is constrained to a single value, then the
implementation of the @b[not] operator is simple@dash()CONSTRAINT_@|ERROR is always
raised; for @b[xor], CONSTRAINT_@|ERROR is always raised if and only if the
component subtype is restricted to the range TRUE .@. TRUE (since TRUE @b[xor]
TRUE yields FALSE and FALSE @b[xor] FALSE yields FALSE).  If the component
subtype range is not known at compile-time, it must be checked at run-time
prior to performing the @b[not] or @b[xor] operation.  In no case is a
component-by-component range check actually required.

No problem arises when operating on scalar values:
@begin(example)
@key[declare]
    @key[type] NB @key[is] @key[new] BOOLEAN @key[range] TRUE .. TRUE;
    A, B : NB := TRUE;
@key[begin]
    @key[if] (@key[not] A) @key[or] (A @key[xor] B) @key[then]     -- no exception is raised
        A := @key[not] A;                  -- CONSTRAINT_@|ERROR
    @key[end] @key[if];
@key[end];
@end(example)
No exception is raised for @b[not] A or A @b[xor] B since the @b[not] and
@b[xor] operators for type NB are the predefined operators, which return
values of the base type.

